/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* --------------------------------------------------- */
/* Functions for multiple precision integer arithmetic */
/* --------------------------------------------------- */

#include        "types.h"
#include        "var.h"

#define         L       BigPrecision    // Some abbreviations
#define         LL      (L+L)           // 2 * BIG_PRECISION

#if     ZZ_ATT
			// No dynamically sized arrays in Cfront
#define         LLL     (MAX_BIG_PRECISION+MAX_BIG_PRECISION)
#else
#define         LLL     LL
#endif

BIG     BIG_0, BIG_1, BIG_INF;          // Constants

static  BIG     max_convertible;        // Max convertible to double

static  int     big_bits;

static  double  powers  [MAX_BIG_PRECISION];
static  int     n_powers;

static  long    bmask [33] = {
				0x00000000,
				0x00000001,
				0x00000003,
				0x00000007,
				0x0000000f,
				0x0000001f,
				0x0000003f,
				0x0000007f,
				0x000000ff,
				0x000001ff,
				0x000003ff,
				0x000007ff,
				0x00000fff,
				0x00001fff,
				0x00003fff,
				0x00007fff,
				0x0000ffff,
				0x0001ffff,
				0x0003ffff,
				0x0007ffff,
				0x000fffff,
				0x001fffff,
				0x003fffff,
				0x007fffff,
				0x00ffffff,
				0x01ffffff,
				0x03ffffff,
				0x07ffffff,
				0x0fffffff,
				0x1fffffff,
				0x3fffffff,
				0x7fffffff,
				0xffffffff      };


void    initBig ()

/* ----------------------------------------- */
/* Initialize the multiple precision package */
/* ----------------------------------------- */

{
double  p, q;
int     i;

	big_bits = L * 31;              // The number of bits in representation

	q = HUGE / 2.0;                 // Double precision powers of two

	for (n_powers = 1, powers [0] = p = 1.0;
		n_powers < L; n_powers++) {

		for (i = 0; i < 31; i++) {
			if (p > q) goto DONE;
			p *= 2.0;
		}

		powers [n_powers] = p;
	}
DONE:
	for (i = 0; i < L; i++) {       // Constants
		BIG_0.x [i] = 0x0;
		BIG_1.x [i] = 0x0;
		BIG_INF.x [i] = 0x7fffffff;
	}

	BIG_1.x [0] = 1;

/* ------------------------------------------------------ */
/* Calculate the maximum BIG number convertible to double */
/* ------------------------------------------------------ */
	for (i = L - 1; i >= n_powers; i--) max_convertible.x [i] = 0;

	for (p = HUGE; i >= 0; i--) {
		q = p / powers [i];
		if (q > powers [1]) {

			if (i == L - 1) {

				max_convertible = BIG_INF;
				break;

			} else {
				cerr << // Quite unlikely to ever get here
				   "Long arithmetic initialization screwup\n";
				exit (17);
			}
		}
		max_convertible.x [i] = (long) q;

		p -= max_convertible.x [i] * powers [i];
	}
}

int  operator< (const BIG &a, const BIG &b) {       // BIG < BIG

	for (int i = L - 1; i >= 0; i--) {
		if (a.x [i] < b.x [i]) return (1);
		if (a.x [i] > b.x [i]) return (0);
	}
	return (0);
}

int  operator<= (const BIG &a, const BIG &b) {      // BIG <= BIG

	for (int i = L - 1; i >= 0; i--) {
		if (a.x [i] > b.x [i]) return (0);
		if (a.x [i] < b.x [i]) return (1);
	}
	return (1);
}

int  operator> (const BIG &a, const BIG &b) {       // BIG > BIG

	for (int i = L - 1; i >= 0; i--) {
		if (a.x [i] > b.x [i]) return (1);
		if (a.x [i] < b.x [i]) return (0);
	}
	return (0);
}

int  operator>= (const BIG &a, const BIG &b) {      // BIG >= BIG

	for (int i = L - 1; i >= 0; i--) {
		if (a.x [i] < b.x [i]) return (0);
		if (a.x [i] > b.x [i]) return (1);
	}
	return (1);
}

BIG  operator+ (const BIG &a, const BIG &b) {       // BIG + BIG

	BIG     r;                      // The result

	for (int i = 0, carry = 0; i < L; i++) {

		if ((r.x [i] = a.x [i] + b.x [i] + carry) < 0) {
			carry = 1;
			r.x [i] &= 0x7fffffff;
		} else {
			carry = 0;
		}
	}

	return (r);
}

BIG  operator+ (const BIG &a, long b) { // BIG + long

	// Note: this function is semantically redundant: c++ could execute
	// BIG + long  as  BIG + (BIG) long.  It exists for improved
	// efficiency

	BIG     r;                      // The result

	if (b >= 0) {

		if ((r.x [0] = a.x [0] + b) >= 0) {     // No carry
			for (int i = 1; i < L; i++) r.x [i] = a.x [i];
			return (r);
		}

		r.x [0] &= 0x7fffffff;                  // Carry

		for (int i = 1; ((r.x [i] = a.x [i] + 1) < 0) && (i < L); i++)
			r.x [i] &= 0x7fffffff;          // Propagate carry

		while (++i < L) r.x [i] = a.x [i];

	} else {

		if ((r.x [0] = a.x [0] + b) >= 0) {       // No carry
			for (int i = 1; i < L; i++) r.x [i] = a.x [i];
			return (r);
		}

		r.x [0] &= 0x7fffffff;

		for (int i = 1; ((r.x [i] = a.x [i] - 1) < 0) && (i < L); i++)
			r.x [i] &= 0x7fffffff;

		while (++i < L) r.x [i] = a.x [i];
	}

	return (r);
}

double  operator+ (const BIG &a, double b) {  // BIG + double (producing double)

	return (b + (double) a);
}

BIG  operator- (const BIG &a, const BIG &b) { // BIG - BIG

	BIG     r;                      // The result

	for (int i = 0, carry = 0; i < L; i++) {

		if ((r.x [i] = a.x [i] - b.x [i] - carry) < 0) {
			carry = 1;
			r.x [i] &= 0x7fffffff;
		} else {
			carry = 0;
		}
	}

	return (r);
}

BIG  operator- (const BIG &a, long b) { // BIG - long/int

	// Redundant -- exists for increased efficiency

	BIG     r;                      // The result

	if (b >= 0) {

		if ((r.x [0] = a.x [0] - b) >= 0) {     // No carry
			for (int i = 1; i < L; i++) r.x [i] = a.x [i];
			return (r);
		}

		r.x [0] &= 0x7fffffff;                  // Carry

		for (int i = 1; ((r.x [i] = a.x [i] - 1) < 0) && (i < L); i++)
			r.x [i] &= 0x7fffffff;          // Propagate carry

		while (++i < L) r.x [i] = a.x [i];

	} else {

		if ((r.x [0] = a.x [0] + b) >= 0) {       // No carry
			for (int i = 1; i < L; i++) r.x [i] = a.x [i];
			return (r);
		}

		r.x [0] &= 0x7fffffff;

		for (int i = 1; ((r.x [i] = a.x [i] + 1) < 0) && (i < L); i++)
			r.x [i] &= 0x7fffffff;

		while (++i < L) r.x [i] = a.x [i];
	}

	return (r);
}

double  operator- (const BIG &a, double b) {  // BIG - double (producing double)

	return ((double) a - b);
}

double  operator- (double a, const BIG &b) {  // double - BIG (producing double)

	return (b - (double) a);
}

BIG  operator* (const BIG &a, const BIG &b) { // BIG * BIG

	// Note: this is slow -- avoid at all cost

	unsigned        long    tmp [LLL];
	BIG             r;              // The result

	for (int i = 0; i < L; i++) tmp [i] = 0;

	r = a * b.x [0];                // Initialize the result

	for (i = L-1; i > 0; i--) {

		if (b.x [i]) {          // Ignore zero factors

			*((BIG*)(&(tmp [L]))) = a * b.x [i];
			r = r + *((BIG*)(&(tmp[L-i])));
		}
	}
	return (r);
}

BIG  operator* (const BIG &a, long b) {  // BIG * long

unsigned        long    aa [LLL], cc [LLL], sc [LLL];
unsigned        long    mp1, mp2, carry, ccc;
int             i, al, k;
BIG             r;                      // The result

	// Convert the operand and calculate the effective length

	for (i = 0, k = 0, al = -1; i < L; i++, k += 2) {
		if ( a.x [i] ) {
			al = k + 2;
			aa [k] = a.x [i] & 0xffff;
			aa [k+1] = (a.x [i] & 0x7fff0000) >> 16;
		} else {
			aa [k] = aa [k+1] = 0;
		}
	}

	if (al <= 0) {                  // The result is 0
		return (BIG_0);
	}

	if (b < 0) {
		return (BIG_0);
	}

	mp1 = b & 0xffff;
	mp2 = (b & 0x7fff0000) >> 16;

	// The first iteration -- initialize the result

	for (i = 0, carry = 0; i < al; ) {

		cc [i] = aa [i] * mp1 + carry;
		carry = (cc [i] & 0xffff0000) >> 16;
		cc [i++] &= 0xffff;

		cc [i] = aa [i] * mp1 + carry;
		carry = (cc [i] & 0xffff8000) >> 15;
		cc [i++] &= 0x7fff;
	}

	while (i < LL) {
		cc [i++] = carry;       carry = 0;
	}

	if (carry) {
		return (BIG_0);
	}

	// The remaining iterations

	if (mp2) {

		for (i = 0, carry = 0; i < al;) {

			aa [i+1] <<= 1;
			if (aa [i] & 0x8000) aa [i+1] |= 1;
			aa [i] &= 0x7fff;

			sc [i] = aa [i] * mp2 + carry;
			carry = (sc [i] & 0xffff8000) >> 15;
			sc [i++] &= 0x7fff;

			sc [i] = aa [i] * mp2 + carry;
			carry = (sc [i] & 0xffff0000) >> 16;
			sc [i++] &= 0xffff;
		}

		if (carry) {

			if (i < LL - 1) {
				if (ccc = ((sc [i] = carry) & 0xffff8000)) {
					sc [i] &= 0x7fff;
					sc [i+1] = ccc >> 15;
					i += 2;
				} else {
					i++;
				}
			} else {
				return (BIG_0);
			}
		}

		for (k = 0, carry = 0; k < i; k++) {

			ccc = sc [k] + carry;
			if ((k >= LL - 1) && ccc) {
				return (BIG_0);
			}

			if (k & 1) {
				carry = ((cc [k+1] += ccc) & 0xffff0000) >> 16;
				cc [k+1] &= 0xffff;
			} else {
				carry = ((cc [k+1] += ccc) & 0xffff8000) >> 15;
				cc [k+1] &= 0x7fff;
			}
		}
	}

	for (i = 0, k = 0; i < L; i++, k += 2)
		r.x [i] = cc [k] | (cc [k+1] << 16);

	return (r);
}

double  operator* (const BIG &a, double b) {  // BIG * double (producing double)

	// Note: the double is supposed to be small enough, so that the result
	// fits into a double.

	int             i;
	double          w, p;
	BIG             sl;

	if (a <= max_convertible) {             // a is convertible to double
		// Convert a to double
		for (i = 0, w = 0.0; i < n_powers; i++)
			if (a.x [i]) w += a.x [i] * powers [i];
		/* And multiply directly */
		return (w * b);
	}

	// Adjust by a power of two (not very efficient but not supposed
	// to be used too often.

	for (i = 0; i < L; i++) sl.x [i] = a.x [i];
	p = 2.0;

	while (1) {
		BShift (sl, -1);
		if (sl <= max_convertible) break;
		p *= 2.0;
	}

	// Now convert BIG operand to double */

	for (i = 0, w = 0.0; i < n_powers; i++)
		if (sl.x [i]) w += sl.x [i] * powers [i];
	return ((w * b) * p);   // Exception possible, but who cares
}

BIG  operator/ (const BIG &a, const BIG &b) {       // BIG / BIG

	// Note: the algorithm for division is naive and slow.  BIG division
	// is not supposed to be used in the simulation. The same about modulo
	// and muliplication of long by long.

	int             pos, i;
	BIG             u, v, rem;
	BIG             r;              // The result

	for (pos = 0; pos < L; pos++)
		if (b.x [pos] != 0) break;
	if (pos >= L) {
		return (BIG_0);
	}
	
	u = b;
	r = BIG_0;
	
	for (pos = 0; ; pos++) {
		if (a < u) {
			pos--;
			BShift (u, -1);
			break;
		}
		if ( u.x [L-1] & 0x40000000 ) break;
		BShift (u, 1);
	}

	for (i = 0; i < L; i++) rem.x [i] = a.x [i];

	if (pos >= 0) {
		while (1) {
			rem = rem - u;
			v.x [0] = 1;
			for (i = 1; i < L; i++) v.x [i] = 0;
			BShift (v, pos);
			r = r + v;
			while (rem < u) {
				--pos;
				BShift (u, -1);
			}
			if (pos < 0) break;
		}
	}
	return (r);
}

BIG  operator% (const BIG &a, const BIG &b) {       // BIG % BIG

	int             pos, i;
	BIG             u, v;
	BIG             r;              // The result

	for (pos = 0; pos < L; pos++)
		if (b.x [pos] != 0) break;
	if (pos >= L) {
		return (BIG_0);
	}
	
	u = b;
	
	for (pos = 0; ; pos++) {
		if (a < u) {
			pos--;
			BShift (u, -1);
			break;
		}
		if ( u.x [L-1] & 0x40000000 ) break;
		BShift (u, 1);
	}

	for (i = 0; i < L; i++) r.x [i] = a.x [i];

	if (pos >= 0) {
		while (1) {
			r = r - u;
			v.x [0] = 1;
			for (i = 1; i < L; i++) v.x [i] = 0;
			BShift (v, pos);
			while (r < u) {
				--pos;
				BShift (u, -1);
			}
			if (pos < 0) break;
		}
	}
	return (r);
}

double  operator/ (const BIG &a, double b) {   // BIG / double (producing double)

	// Note: the double is supposed to be big enough, so that the result
	// fits into a double.

	int             i;
	double          w, p;
	BIG             sl;

	if (a <= max_convertible) {     // Convert a to double
		for (i = 0, w = 0.0; i < n_powers; i++)
			if (a.x [i]) w += a.x [i] * powers [i];
					// And divide directly
		return (w / b);
	}

	// Adjust by a power of two

	for (i = 0; i < L; i++) sl.x [i] = a.x [i];
	p = 2.0;

	while (1) {
		BShift (sl, -1);
		if (sl <= max_convertible) break;
		p *= 2.0;
	}

	// Now convert BIG operand to double

	for (i = 0, w = 0.0; i < n_powers; i++)
		if (sl.x [i]) w += sl.x [i] * powers [i];
	return ((w / b) * p);
}

double  operator/ (double a, const BIG &b) {   // double / BIG (producing double)

	int             i;
	double          w, p;
	BIG             sl;

	if (b <= max_convertible) {     // Convert a to double
		for (i = 0, w = 0.0; i < n_powers; i++)
			if (b.x [i]) w += b.x [i] * powers [i];
					// And divide directly
		return (a / w);
	}

	// Adjust by a power of two

	for (i = 0; i < L; i++) sl.x [i] = b.x [i];
	p = 2.0;

	while (1) {
		BShift (sl, -1);
		if (sl <= max_convertible) break;
		p *= 2.0;
	}

	// Now convert BIG operand to double

	for (i = 0, w = 0.0; i < n_powers; i++)
		if (sl.x [i]) w += sl.x [i] * powers [i];
	return ((a / w) / p);
}

long  operator% (const BIG &a, long b) { // BIG % long (producing long)

	BIG     r;

	r = a % ((BIG) b);

	return (r.x [0]);
}

BIG::BIG (double a) {                   // Constructor from double

	int     i;
	double  f;
	
	if (a < 0.0) {
		*this = BIG_0;
	}

	for (i = L-1; i >= 0; i--) {
		if (i >= n_powers) {
			x [i] = 0;
			continue;
		}
		f = a / powers [i];
		if (f >= powers [1]) {
			*this = BIG_INF;
		}
		x [i] = (long) f;
		a -= ((double) x [i]) * powers [i];
	}
}

/*
long    BIG::operator  long () {                // Conversion to long

	return (x [0]);
}
*/

BIG::operator  double () const {                // Conversion to double

	int     i;
	double  d;

	if (*this > max_convertible) {
		return (HUGE);
	}
	for (i = 0, d = 0.0; i < n_powers; i++)
		if (x [i]) d += x [i] * powers [i];
	return (d);
}

int  convertible (const BIG &a) {       // Checks convertibility to double

	return (a <= max_convertible);
}

BIG  ctob (char *s) {                   // String to BIG conversion

	BIG     r;

	while (*s == ' ') s++;
	if (*s == '+') s++;
	while (*s == ' ') s++;

	r = BIG_0;

	while ((*s <= '9') && (*s >= '0')) {
		r = r * 10 + (*s - '0');
		s++;
	}

	return (r);
}

char    *btoa (const BIG &a, char* s, int nc) { // BIG to string

	unsigned        long    aa [LLL];
	unsigned        long    carry, ccc;
	int             i, j, al, k;
	char            temp [256], *tm, *res;  // Hopefully enough

	/* Convert the operand and calculate the effective length */

	for (res = s, i = 0, k = 0, al = -1; i < L; i++, k += 2) {
		if ( a.x [i] ) {
			al = k + 2;
			aa [k] = a.x [i] & 0xffff;
			aa [k+1] = (a.x [i] & 0x7fff0000) >> 16;
		} else {
			aa [k] = aa [k+1] = 0;
		}
	}

	tm = &(temp [255]);

	for (j = 0, al--; j < 256; j++) {

		if (al < 0) {
			if (j == 0) *--tm = '0'; else break;
			continue;
		}

		for (i = al, al = -1; i >= 0; ) {

			carry = aa [i] - (ccc = aa [i] / 10) * 10;
			if ((aa [i] = ccc) && (al < 0)) al = i;
			aa [--i] |= (carry << 16);
			carry = aa [i] - (ccc = aa [i] / 10) * 10;
			if ((aa [i] = ccc) && (al < 0)) al = i;
			if (--i >= 0)
				aa [i] |= (carry << 15);
		}

		if (!(al & 1)) al++;

		*--tm = '0' + (char) carry;
	}

	if (j <= nc) {
		while (nc-- >  j) *s++ = ' ';
		while (nc-- >= 0) *s++ = *tm++;
		*s = '\0';
		return (res);
	}

	if ((i = j - nc + 2) > 9)
		i = j - nc + 3;

	if (i >= j) {
		while (nc-- > 0) *s++ = '*';
		*s = '\0';
		return (res);
	}

	while (j-- > i) *s++ = *tm++;
	*s++ = 'E';
	if (i > 9) {
		*s++ = (j = (i/10)) + '0';
		*s++ = (i - (j * 10)) + '0';
	} else {
		*s++ = i + '0';
	}
	*s ='\0';
	return (res);
}

void    BShift (BIG &a, int c) {        // Shifts BIG by c binary positions

	int     w, b, i;
	long    p, q, m;

	if (c > 0) {
		if (c >= big_bits) {
			for (i = 0; i < L; i++) a.x [i] = 0;
			return;
		}
		w = c / 31;             /* Words */
		b = c % 31;             /* Bits */

		if (w) {
			for (i = L - 1; i >= w; i--)
				a.x [i] = a.x [i - w];
			for (; i >= 0; i--) a.x [i] = 0;
		}

		if (b) {
			m = bmask [b] << (31 - b);
			for ( i = w, p = 0; i < L; i++) {
				q = (a.x [i] & m) >> (31 - b);
				a.x [i] = ((a.x [i] & ~m) << b) | p;
				p = q;
			}
		}
	} else {
		if (c <= -big_bits) {
			for (i = 0; i < L; i++) a.x [i] = 0;
			return;
		}
		w = (c = -c) / 31;      // Words
		b = c % 31;             // Bits

		if (w) {
			for (i = 0; i < L - w; i++)
				a.x [i] = a.x [i + w];
			for (; i < L; i++) a.x [i] = 0;
		}

		if (b) {
			m = bmask [b];
			for ( i = L - w - 1, p = 0; i >= 0; i--) {
				q = (a.x [i] & m) << (31 - b);
				a.x [i] = ((a.x [i] & ~m) >> b) | p;
				p = q;
			}
		}
	}
}

/* ---------------------------------------- */
/* ATT Cfront doesn't like loops in inlines */
/* ---------------------------------------- */

#if     ZZ_ATT
BIG::BIG (int i) {

/* ------------------- */
/* Conversion from int */
/* ------------------- */

	x [0] = (long) i;
	for (int k = 1; k < BigPrecision; x [k++] = 0L);
};

BIG::BIG (long i) {

/* -------------------- */
/* Conversion from long */
/* -------------------- */

	x [0] = i;
	for (int k = 1; k < BigPrecision; x [k++] = 0L);
};

int  operator== (const BIG &a, const BIG &b) {
	for (int k = 0; k < BigPrecision; k++)
		if (a.x [k] != b.x [k]) return (0);
	return (1);
};

int  operator!= (const BIG &a, const BIG &b) {
	for (int k = 0; k < BigPrecision; k++)
		if (a.x [k] != b.x [k]) return (1);
	return (0);
};
#endif
