/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* -------------------------- */
/* Internal display functions */
/* -------------------------- */

#include        <signal.h>
#include        <setjmp.h>
#include        "system.h"
#include        "display.h"

/* --------------------------- */
/* Macros to set timeout clock */
/* --------------------------- */
#define         SETTIMER                zz_settimer (zz_ctimeout, zz_tmoutenv)
#define         CLEARTIMER              zz_cleartimer ();

/* ------------------------------------------------------ */
/* Macros to set/reset asynchronous input from the server */
/* ------------------------------------------------------ */
#define         SYNCIO                  zz_setsyncio ();
#define         ASYNCIO                 zz_setasyncio ((SIGARG)iosig);

/* --------------------- */
/* Same from the monitor */
/* --------------------- */
#define         MSTOP                   zz_monitorstop ();
#define         MWAIT                   zz_monitorwait ((SIGARG)monitorRequest);

/* ------------------------ */
/* Some function prototypes */
/* ------------------------ */
int     zz_getByteFromServer (), zz_getByteFromMonitor ();
void    zz_sendByteToServer (int), zz_sendTextToMonitor (char*);
int     zz_connectToMonitor (), zz_openLine ();
void    zz_closeLine ();
void    zz_setsyncio (), zz_setasyncio (SIGARG);
void    zz_monitorstop (), zz_monitorwait (SIGARG);
void    zz_cleartimer ();
void    zz_settimer (long, jmp_buf);
static  void    monitorRequest ();

jmp_buf         zz_tmoutenv;            // Timeout exception environment for
					// setjmp

static  volatile int     initializing,  // Flag = initial handshake
			 withinwindow;  // True if within expose

static  int     itemcount,              // Limit for the number of wind. items
		peekoctet,              // Current byte read from server
		sendindicator;          // Tells if something was sent to server

static  char    *rplyptr;               // Reply pointer for RPL

ZZ_WINDOW::ZZ_WINDOW (ZZ_Object *o, int md, int stat, int cnt) {

/* --------------------------------- */
/* Constructor for window list items */
/* --------------------------------- */

	obj = o;
	mode = md;
	flag = WF_OPENING;              // Newly-created window
	stid = stat;
	count = cnt;
	frame = NULL;                   // No associated object
}

static  void    iosig () {

/* ---------------------------------------------------------- */
/* I/O  signal  processor  for  asynchronous  input  from the */
/* server                                                     */
/* ---------------------------------------------------------- */

	zz_events_to_display = 0;
	return;
}

static  int     getOctet () {

/* ------------------------------ */
/* Reads one byte from the server */
/* ------------------------------ */

	int     r;

	if (peekoctet != NONE) {
		// Never get past ESP
		if ((r = peekoctet) != PH_ESP) peekoctet = NONE;
		return (r);
	}

	if ((r = zz_getByteFromServer ()) == PH_ESP) peekoctet = PH_ESP;
	return (r);
}

static  int     peekOctet () {

/* ---------------------------------------------------------- */
/* Look  at  the  next octet arriving from the server without */
/* reading it                                                 */
/* ---------------------------------------------------------- */

	if (peekoctet == NONE) peekoctet = zz_getByteFromServer ();
	return (peekoctet);
}

inline  static  void    sendOctet (int val) {

/* --------------------------- */
/* Send one byte to the server */
/* --------------------------- */

	zz_sendByteToServer (val);
}

static  int     dscrhalf, dscrlast;

inline  static  int     getDescriptor () {

/* --------------------------------------------------- */
/* Gets one numeric descriptor from the input sequence */
/* --------------------------------------------------- */

	register int    rslt;

	if (dscrhalf) {
		// Take the second half
		dscrhalf = NO;
		rslt = dscrlast & 017;
	} else {
		// Get a new byte
		dscrlast = getOctet ();
		dscrhalf = YES;
		rslt = (dscrlast >> 4) & 017;
	}
	return (rslt);
}

static  void    sendText (char *t) {

/* ------------------------------ */
/* Send text string to the server */
/* ------------------------------ */

	int     l;

	sendOctet (PH_TXT);
	while (*t != '\0')
		sendOctet ((l = *t++) & 0377);
	sendOctet (0);
}

static  void    sendError (int code, char *t) {

/* --------------------- */
/* Sends an error phrase */
/* --------------------- */

	sendOctet (PH_ERR);
	sendOctet (code);
	sendText (t);
}

static  long    getInt () {

/* --------------------------------------------- */
/* Reads an integer number from the input stream */
/* --------------------------------------------- */

	long    res;
	int     des;

	// This shouldn't really happen
	if (peekOctet () != PH_INM) return (0);
	getOctet ();

	res = 0; dscrhalf = NO;
	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP)
			// Positive number
			return (-res);
		else if (des == DSC_ENN)
			// Negative number
			return (res);
		else if (des > 9) {
			// Spurt error -- not a decimal digit
			while (getOctet () != PH_ESP);
			sendError (DERR_ILS, "Illegal integer number");
			sendindicator = YES;
			return (0);
		}
		res = res * 10 - des;
	}

	return (res);
}

static  BIG     getBIG () {

/* ---------------------------------------- */
/* Reads a BIG number from the input stream */
/* ---------------------------------------- */

	BIG     res;
	int     des;

	// This shouldn't really happen
	if (peekOctet () != PH_BNM) return (BIG_0);
	getOctet ();

	res = BIG_0; dscrhalf = NO;
	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP)
			// Positive number
			return (res);
			// Note: no negative big numbers allowed
		else if (des > 9) {
			// Spurt error -- not a decimal digit
			while (getOctet () != PH_ESP);
			sendError (DERR_ILS, "Illegal BIG number");
			sendindicator = YES;
			return (BIG_0);
		}
		res = res * 10 + des;
	}

	return (res);
}

#if 0
	// Not used yet
static  double  getFloat () {

/* --------------------------------------------------- */
/* Reads a floating point number from the input stream */
/* --------------------------------------------------- */

	double  res, fac, frac;
	int     des, exp;

	if (peekOctet () != PH_FNM) return (0.0);
	getOctet ();

	res = 0.0; dscrhalf = NO;

	// Before dot

	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP)
			break;
		else if (des == DSC_DOT)
			goto GF_DOT;
		else if (des == DSC_ENN) {
			res = -res;
			break;
		} else if (des > 9) {
			// Spurt error -- not a decimal digit
GF_SPERR:
			while (getOctet () != PH_ESP);
			sendError (DERR_ILS, "Illegal FP number");
			sendindicator = YES;
			return (0.0);
		}
		res = res * 10.0 + (double) des;
	}

	// Now for the fractional part

	if (getDescriptor () != DSC_DOT) goto GF_SPERR;
GF_DOT:
	fac = 1.0;
	frac = 0.0;

	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP)
			break;
		else if (des == DSC_EXP)
			goto GF_EXP;
		else if (des == DSC_ENN) {
			frac = -frac;
			break;
		} else if (des > 9) goto GF_SPERR;

		frac = frac + (((double) des) * (fac *= 0.1));
	}

	// Now for the exponent

	if (getDescriptor () != DSC_EXP) goto GF_SPERR;
GF_EXP:
	res += frac;
	exp = 0;

	while (1) {
		if ((des = getDescriptor ()) == DSC_ENP)
			break;
		else if (des == DSC_ENN) {
			exp = -exp;
			break;
		} else if (des > 9) goto GF_SPERR;

		exp = exp * 10 + des;
	}

	fac = pow (10.0, exp);
	res *= fac;
	return (res);
}
#endif

static  int     lastsdscr, lastshalf;

inline  static  void    sendDescriptor (int d) {

/* ----------------------------- */
/* Send one numerical descriptor */
/* ----------------------------- */

	if (lastshalf) {
		sendOctet (lastsdscr | (d & 017));
		lastshalf = NO;
	} else {
		lastshalf = YES;
		lastsdscr = (d & 017) << 4;
	}
}

inline  static  void    flushDescriptor () {

/* ------------------------------------------------------ */
/* Flush a partially filled descriptor byte to the server */
/* ------------------------------------------------------ */

	if (lastshalf) {
		sendOctet (lastsdscr);
		lastshalf = NO;
	}
}

static  void    sendInt (long val) {

/* ------------------------------------ */
/* Send an integer number to the server */
/* ------------------------------------ */

	int     digits [16], ndigits, sig;

	sendOctet (PH_INM);

	lastshalf = NO;

	if (val >= 0) {
		sig = NO;
		val = -val;
	} else
		sig = YES;

	for (ndigits = 0; val; ) {
		digits [ndigits++] = - (int)(val % 10);
		val /= 10;
	}

	while (ndigits--) {
		sendDescriptor (digits [ndigits]);
	}
	sendDescriptor (sig ? DSC_ENN : DSC_ENP);
	flushDescriptor ();
}

static  void    sendBIG (BIG val) {

/* ------------------------------- */
/* Send a BIG number to the server */
/* ------------------------------- */

	int     digits [256], ndigits;

	sendOctet (PH_BNM);

	lastshalf = NO;

	for (ndigits = 0; val != BIG_0; ) {
		digits [ndigits++] = (int)(val % 10);
		val /= 10;
	}

	while (ndigits--) sendDescriptor (digits [ndigits]);
	sendDescriptor (DSC_ENP);
	flushDescriptor ();
}

static  void    sendFloat (double val) {

/* ------------------------------------------ */
/* Send a floating point number to the server */
/* ------------------------------------------ */

	char    *digits;
	int     sign;
#ifdef  ZZ_OLB
	char    tmp [32];               // Do not use 'dtoa'
#endif

	sendOctet (PH_FNM);
	lastshalf = NO;
#ifdef  ZZ_OLB
	sprintf (tmp, "%g", val);
	digits = tmp;
#else
	digits = dtoa (val);
#endif
	sign = NO;

	while (*digits != '\0') {
		if (*digits == ' ' || *digits == '+') {
			digits++;
			continue;
		}

		if (*digits == '-') {
			sign = YES;
			digits++;
			continue;
		}

		if (*digits == '.' || *digits == 'e' || *digits == 'E') {
			// No leading part
			sendDescriptor (DSC_ENP);
			sendDescriptor (DSC_DOT);
			if (*digits != '.') {
				sendDescriptor (DSC_ENP);
				sendDescriptor (DSC_EXP);
				sendDescriptor (DSC_ENP);
				flushDescriptor ();
				return;
			}
			break;          // Go process fractional part
		}

		// Process leading part (we assume that the dtoa output
		// is sane

		while (*digits <= '9' && *digits >= '0')
			sendDescriptor ((*digits++) - '0');
		sendDescriptor (sign ? DSC_ENN : DSC_ENP);
		sendDescriptor (DSC_DOT);
		break;
	}

	// The integer part is done

	if (*digits == '\0') {
		sendDescriptor (DSC_ENP);
		sendDescriptor (DSC_EXP);
		sendDescriptor (DSC_ENP);
		flushDescriptor ();
		return;
	}

	if (*digits == '.') {
		// Process fractional part
		digits++;
		while (*digits <= '9' && *digits >= '0')
			sendDescriptor ((*digits++) - '0');
		sendDescriptor (sign ? DSC_ENN : DSC_ENP);
	} else {
		// Null fractional part
		sendDescriptor (DSC_ENP);
	}

	sendDescriptor (DSC_EXP);

	// Now the exponent

	if (*digits != 'e' && *digits != 'E') {
		sendDescriptor (DSC_ENP);
		flushDescriptor ();
		return;
	}

	digits++;

	sign = NO;

	while (*digits != '\0') {
		if (*digits == '+') {
			digits++;
			continue;
		}
		if (*digits == '-') {
			sign = YES;
			digits++;
			continue;
		}
		while (*digits <= '9' && *digits >= '0')
			sendDescriptor ((*digits++) - '0');
		break;
	}
	sendDescriptor (sign ? DSC_ENN : DSC_ENP);
	flushDescriptor ();
}

static  void    serviceEND () {         // Process END phrase

	ZZ_WSIT    *wst;
	ZZ_SIT     *st;
	ZZ_WINDOW  *w;
	ZZ_SUPPLI  *supi;
	ZZ_SUPPLD  *supd;

	zz_closeLine ();

	zz_ctimeout = CTIMEOUT;         // Assume default connection timeout
	DisplayActive = NO;
	DisplayInterval = DEFDINTVL;
	zz_send_step_phrase = NO;

	// Deallocate lists

	withinwindow = YES;
	for (zz_the_window = zz_windows; zz_the_window != NULL; ) {
		w = zz_the_window->next;
		DisplayClosing = YES;
		itemcount = 0;
		DisplayOpening = NO;
		if (zz_the_window->frame != NULL)
			zz_the_window->obj->displayOut (zz_the_window->mode,
				zz_the_window->stid);
		delete (zz_the_window);
		zz_the_window = w;
	}
	withinwindow = NO;

	while (zz_steplist != NULL) {
		st = zz_steplist->next;
		delete (zz_steplist);
		zz_steplist = st;
	}

	while (zz_steplist_delayed != NULL) {
		wst = zz_steplist_delayed->next;
		delete (zz_steplist_delayed);
		zz_steplist_delayed = wst;
	}

	while (zz_supplist != NULL) {
		supi = zz_supplist->next;
		delete (zz_supplist);
		zz_supplist = supi;
	}

	while (zz_suppldst != NULL) {
		supd = zz_suppldst->next;
		delete (zz_suppldst);
		zz_suppldst = supd;
	}

	zz_windows = NULL;              // Active windows
	zz_steplist = NULL;             // Stepped windows
	zz_steplist_delayed = NULL;     // Waiting to be stepped
	zz_supplist = NULL;
	zz_suppldst = NULL;
	zz_suppcount = 0;
	zz_commit_event = MAX_long;
	zz_commit_time = TIME_inf;

	initializing = NO;

	// Resume monitor wait mode
	MWAIT;
}

static  void    monitorRequest () {

/* ----------------------- */
/* Process monitor request */
/* ----------------------- */

	// Note: this junk is executed as an interrupt that can arrive at
	//       any moment. Make sure that you do not:
	//
	//       -- assume the consistency of any lists, etc.
	//       -- do anything that may affect the consistency of
	//          partially executed operations (e.g. memory allocation
	//          is forbidden)

	MSTOP;                          // Stop waiting for the monitor

	if (setjmp (zz_tmoutenv)) {
		// After timing out on monitor I/O
		CLEARTIMER;
		MWAIT;
		return;                 // Ignore
	}

	SETTIMER;

	switch (zz_getByteFromMonitor ()) {

	  // The first (and hopefully the only) byte identifies the request

	  case PH_RPL:                  // Status request

		CLEARTIMER;

		{
			char    txt [16];

			SETTIMER;
			sprintf (txt, "%15d", zz_NRMessages);
			// 'form' is an example of an unsafe operation (it
			// affects the status of a buffer pool). This is why
			// sprintf is used instead
			zz_sendTextToMonitor (txt);
			btoa (Time, txt, 15);
			zz_sendTextToMonitor (txt);
			sprintf (txt, "%15.3f", cpuTime ());
			zz_sendTextToMonitor (txt);
			CLEARTIMER;
			MWAIT;
			return;
		}

	  case PH_ATT:                  // Display request

		CLEARTIMER;

		{
			if (setjmp (zz_tmoutenv)) {
				// Closing after timeout. Note: calling
				// serviceEND is safe as display routines
				// are never active in the MWAIT mode
				CLEARTIMER;
				serviceEND ();
				return;
			}
				
			SETTIMER;

			if (zz_openLine () < 0) {
				CLEARTIMER;
				serviceEND ();
				return;
			}

			CLEARTIMER;

			// Do not resume the monitor wait mode. Do not respond
			// to the forthcoming monitor request while in display
			// mode

			DisplayActive = YES;
			// Flag initial handshake
			initializing = YES;
			// Force immediate attention of smurph
			zz_events_to_display = 0;
			withinwindow = NO;
			return;
		}

	  default:

		CLEARTIMER;             // Urecognizable -- ignore
		MWAIT;
	}
}

void    zz_init_windows () {            // Initialize display module

	int     f;

	zz_ctimeout = CTIMEOUT;         // Default connection timeout

	if ((f = zz_connectToMonitor ()) >= 0) MWAIT;

	if (zz_drequest) {              // Requested by '-d'
		if (f < 0)
			excptn ("initWindows: cannot connect to monitor");
		zz_events_to_display = -1;
		// Wait for the display server
		while (zz_events_to_display) sigpause (0);
	}
	zz_drequest = f;                // If < 0 display impossible
}

int     requestDisplay (char *msg) {

/* -------------------------------- */
/* Display request from the program */
/* -------------------------------- */

	if (zz_drequest < 0)
excptn ("requestDisplay: display impossible -- no connection to monitor");
	zz_ctimeout = UTIMEOUT;

	if (DisplayActive) return (ERROR);   // Display already active

	zz_events_to_display = -1;
	// Wait for the display server
	while (zz_events_to_display) sigpause (0);
	if (msg != NULL) displayNote (msg);
	return (OK);
}

static  char    *stchp, *stchq;         // Two scratch variables for sendTree

void zz_sendTree (ZZ_Object *o) {

/* ---------------------------------------------------------- */
/* Sends  to  the  server  the  initial  ownership  tree as a */
/* sequence of updates to System                              */
/* ---------------------------------------------------------- */

	ZZ_Object  *clist, *osent;

	switch (o->Class) {

		case    OBJ_station:

			clist = ((Station*)o)->ChList;
			osent = (ZZ_Object*) (&((Station*)o)->ChList);
			break;

		case    AIC_traffic:

			clist = ((Traffic*)o)->ChList;
			osent = (ZZ_Object*) (&((Traffic*)o)->ChList);
			break;

		case    AIC_process:

			clist = ((Process*)o)->ChList;
			osent = (ZZ_Object*) (&((Process*)o)->ChList);
			break;

		default:

			return;         // Nothing to do
	}

	while (clist != NULL) {

	    if (!zz_flg_nosysdisp || clist->Class != AIC_process ||
			((Process*)clist)->Owner != System || clist == Kernel
				|| clist == ZZ_Main) {

		// Ignore system processes, if '-s' hasn't been selected

		sendOctet (PH_ADD);

		// The parents standard name
		sendText (o->getSName ());

		// The object's standard name
		sendText (stchp = clist->getSName ());

		// The object's base name
		if (strcmp (stchp, stchq = clist->getBName ()) == 0)
			sendText ("");
		else
			sendText (stchq);

		// The object's nickname
		if (clist->zz_nickname != NULL)
			sendText (clist->zz_nickname);
		else
			sendText ("");

		sendOctet (PH_EPH);

		zz_sendTree (clist);    // Recursively for the child
	    }
	    clist = clist->next;

	}
}

static  char       *objtofind;          // To make recursion less expensive
static  ZZ_Object  *objfound, *fchild;

ZZ_Object *zz_findAdvance (ZZ_Object *o) {

/* --------------------------------- */
/* Recursive searcher for findObject */
/* --------------------------------- */

	ZZ_Object  *clist;

	if (strcmp (o->getSName (), objtofind) == 0) return (o);

	switch (o->Class) {

		case    OBJ_station:

			clist = ((Station*)o)->ChList;
			break;

		case    AIC_traffic:

			clist = ((Traffic*)o)->ChList;
			break;

		case    AIC_process:

			clist = ((Process*)o)->ChList;
			break;

		default:

			return (NULL);          // Nothing to do
	}

	if (clist == NULL) return (NULL);       // Same thing

	for ( ; clist != NULL; clist = clist->next)
		if ((objfound = zz_findAdvance (clist)) != NULL)
			return (objfound);

	return (NULL);                          // Not found there
}

ZZ_Object *zz_findWAdvance (ZZ_Object *o) {

/* -------------------------------- */
/* Recursive searcher for findOwner */
/* -------------------------------- */

	ZZ_Object  *clist;

	switch (o->Class) {

		case    OBJ_station:

			clist = ((Station*)o)->ChList;
			break;

		case    AIC_traffic:

			clist = ((Traffic*)o)->ChList;
			break;

		case    AIC_process:

			clist = ((Process*)o)->ChList;
			break;

		default:

			return (NULL);          // Nothing to do
	}

	if (clist == NULL) return (NULL);       // Same thing

	for ( ; clist != NULL; clist = clist->next) {
		if (clist == fchild) return (o);
		if ((objfound = zz_findWAdvance (clist)) != NULL)
			return (objfound);
	}

	return (NULL);                          // Not found there
}

static  ZZ_Object  *findOwner (ZZ_Object *o) {

/* ----------------------------- */
/* Finds the owner of object 'o' */
/* ----------------------------- */

	fchild = o;
	return (zz_findWAdvance (System));

}

static  ZZ_Object  *findObject (char *sn) {

/* ---------------------------------------------------------- */
/* Searches  the  system  object tree for the object with the */
/* given standard name                                        */
/* ---------------------------------------------------------- */

	objtofind = sn;

	return (zz_findAdvance (System));

}

void  zz_processWindowPhrase () {

/* ------------------------------------------------ */
/* Reads one window phrase arriving from the server */
/* ------------------------------------------------ */

	char    sname [256];    // The object's standard name
	int     i, orgmode,
		mode,           // Window mode
		stat;           // Station relative

	BIG     steptime;
	long    stepevent;
	int     delmode;        // Defer mode

	int     count;          // Item count limit

	ZZ_WINDOW  *w;          // Window list pointers
	ZZ_SIT     *ws;
	ZZ_WSIT    *wsd;

	Station         *stepstation;   // Step parameters
	Process         *stepprocess;
	AI              *stepai;
	Observer        *stepobs;

	// Get the standard name of the object

	if (getOctet () != PH_TXT) {
		// Formal error
		while (getOctet () != PH_ESP);  // Skip the rest of the spurt
		sendError (DERR_ILS, "Standard name missing from WND phrase");
		sendindicator = YES;
		return;
	}

	for (i = 0; i < 256; i++)
		if ((sname [i] = getOctet ()) == NULL) break;

	if (i >= 256) {
		// Standard name too long -- assume "illegal spurt" error
		while (getOctet () != PH_ESP);  // Skip the rest
		sendError (DERR_ILS,
		"Standard name in WND phrase is longer than 256 characters");
		sendindicator = YES;
		return;
	}

	if (peekOctet () != PH_INM) {
		// Error -- mode indicator expected
		while (getOctet () != PH_ESP);  // Skip the rest
		sendError (DERR_ILS, "Mode indicator missing in a WND phrase");
		sendindicator = YES;
		return;
	}

	if ((orgmode = (int) getInt ()) < 0) {
		// Illegal mode. The case of a too high mode cannot be detected
		// here. It will be detected dynamically when the window is
		// displayed.
		mode = MAX_int; // Force it to be positive, but still illegal
	} else
		mode = orgmode;

	// Check if the window is to be station-relative

	if (peekOctet () == PH_INM) {
		// Station-relative window
		if ((stat = (int) getInt ()) < 0 || stat > NStations - 1) {
			// Illegal station -- skip to the end of the window
			// phrase
			sendindicator = YES;
			if (peekOctet () == PH_STP) {
				// Skip the STP phrase
				getOctet ();
				if ((i = peekOctet ()) == PH_BNM) {
					while (peekOctet () == PH_EPH)
					getBIG ();
						getOctet ();
				} else if (i == PH_INM) {
					getInt ();
					while (peekOctet () == PH_EPH)
						getOctet ();
				} else if (i != PH_ESM) getOctet ();
				if (peekOctet () == PH_REL) getOctet ();
			}
			if (peekOctet () == PH_TRN) {
				getOctet ();
				if (getOctet () != PH_INM) {
					// Formal error
					while (getOctet () != PH_ESP);
					sendError (DERR_ILS,
						"TRN subphrase incorrect");
					return;
				}
				getInt ();
			}

			// Send the "illegal station" message
			sendOctet (PH_WCN);
			sendText (sname);
			sendInt (orgmode);
			sendInt (stat);
			sendError (DERR_ISN, "Illegal station number");
			sendOctet (PH_EPH);
			return;
		}
	} else
		stat = NONE;    // Not station-relative

	stepevent = MAX_long;
	steptime = BIG_inf;

	if (peekOctet () == PH_STP) {
		getOctet ();
		// Process STP subphrase
		if ((i = peekOctet ()) == PH_BNM) {
			// Time-delayed
			delmode = 1;
			steptime = getBIG ();
			if (peekOctet () == PH_REL) {
				// Relative
				steptime += Time;
				getOctet ();
			}
		} else if (i == PH_INM) {
			// Event-delayed
			delmode = -1;
			stepevent = getInt ();
			if (peekOctet () == PH_REL) {
				// Relative
				stepevent += zz_npre;
				getOctet ();
			}
		} else if (i == PH_ESM) {       // Exit step mode
			delmode = 2;
			getOctet ();
		} else
			delmode = 0;            // Undelayed step

	} else
		delmode = -2;                   // No step mode

	// Check if item limit specified

	count = MAX_int;
	if (peekOctet () == PH_TRN) {
		getOctet ();
		if (peekOctet () != PH_INM) {
			// Formal error
			while (getOctet () != PH_ESP);
			sendError (DERR_ILS, "TRN phrase corrupted");
			sendindicator = YES;
			return;
		}
		if ((count = (int) getInt ()) < 0) count = 0;
	}

	// The entire phrase has been read in. Now process it

	if (delmode == 2) {

		// Exit step mode

		CLEARTIMER;

		for (ws = zz_steplist; ws != NULL; ws = ws->next) {
			// Find the window on the step list
			if (ws->win->mode != mode ||
				ws->win->stid != stat) continue;
			if (strcmp (ws->win->obj->getSName (), sname) != 0)
				continue;
			// We've got it

			queue_out (ws);
			delete (ws);
			break;
		}

		if (ws != NULL) {
			SETTIMER;
			return;
		}

		// Not found there: check if deferred

		for (wsd = zz_steplist_delayed; wsd != NULL; wsd = wsd->next) {
			// Find the window on the step list
			if (wsd->win->mode != mode ||
				wsd->win->stid != stat) continue;
			if (strcmp (wsd->win->obj->getSName (), sname) != 0)
				continue;
			// We've got it

			queue_out (wsd);
			delete (wsd);

			// Perhaps we should update here the minimum intervals
			// until the earliest deferred window will be stepped.
			// We don't do it. The event manager in main will have
			// to check whether such a window exists and if not,
			// update the intervals appropriately.

			break;
		}

		SETTIMER;

		if (wsd == NULL) {      // Window not found
			sendOctet (PH_WCN);
			sendText (sname);
			sendInt (orgmode);
			sendInt (stat);
			sendError (DERR_CUN, "No window to unstep");
			sendOctet (PH_EPH);
			sendindicator = YES;
		}

		return;                 // This part done
	}

	// Add the window, if not present

	for (w = zz_windows; w != NULL; w = w->next) {
		// Check if the window occurs on the active list
		if (w->mode != mode || w->stid != stat) continue;
		if (strcmp (w->obj->getSName (), sname) == 0) break;
	}

	if (w == NULL) {                // Must be added

		ZZ_Object  *o;

		// Find the object in question
		if ((o = findObject (sname)) == NULL) {
			// No such object
			sendOctet (PH_WCN);
			sendText (sname);
			sendInt (orgmode);
			sendInt (stat);
			sendError (DERR_ONF, "Object not found");
			sendOctet (PH_EPH);
			sendindicator = YES;
			return;
		}

		CLEARTIMER;
		w = new ZZ_WINDOW (o, mode, stat, count);
		queue_head (w, zz_windows, ZZ_WINDOW);
		SETTIMER;

	} else {

		// A possible count modification

		if (count != MAX_int)
			w->count = count;
	}

	// Now determine the step status

	if (delmode == -2) return;      // No stepping

	if (w->flag & WF_OPENING) w->flag |= WF_STEPPED;

	// Step the window

	stepstation     = NULL;
	stepprocess     = NULL;
	stepai          = NULL;
	stepobs         = NULL;

	switch (w->obj->Class) {

	  ZZ_Object *o;

	  case  OBJ_station:

		// Step this station and nothing else
		stepstation = (Station*)(w->obj);
		break;

	  case  AIC_process:

		// Step this process
		if (w->obj == Kernel) break;    // Global step
		stepprocess = (Process*)(w->obj);
		break;

	  case  OBJ_rvariable:

		o = findOwner (w->obj);
		Assert (o != NULL,
		      "processWindowPhrase: internal error -- owner not found");
		if (o->Class == OBJ_station)
			stepstation = (Station*)o;
		else if (o->Class == AIC_process)
			stepprocess = (Process*)o;
		else if (o->Class == AIC_traffic)
			stepai = (AI*)o;
		else
    excptn ("processWindowPhrase: internal error -- illegal RVariable owner");

		break;

	  case  OBJ_eobject:

		o = findOwner (w->obj);
		Assert (o != NULL,
		      "processWindowPhrase: internal error -- owner not found");
		if (o->Class == OBJ_station)
			stepstation = (Station*)o;
		else if (o->Class == AIC_process)
			stepprocess = (Process*)o;
		else
    excptn ("processWindowPhrase: internal error -- illegal EObject owner");

		break;

	  case  AIC_observer:

		stepobs = (Observer*) (w->obj);
		break;
	
	  default:

		// Only AIs are left
		stepai = (AI*)(w->obj);
		if (w->obj->Class != AIC_port && stat != NONE)
			stepstation = idToStation (stat);
	}

	CLEARTIMER;
	if (delmode == 0) {             // Immediate stepping

		ws = new ZZ_SIT;
		ws->win = w;            // For this window
		ws->station = stepstation;
		ws->process = stepprocess;
		ws->ai      = stepai;
		ws->obs     = stepobs;

		queue_head (ws, zz_steplist, ZZ_SIT);
		SETTIMER;
		return;
	}

	// Deferred stepping

	wsd = new ZZ_WSIT;
	wsd->win = w;                   // For this window
	wsd->station = stepstation;
	wsd->process = stepprocess;
	wsd->ai      = stepai;
	wsd->obs     = stepobs;
	wsd->cevent  = stepevent;
	wsd->ctime   = steptime;

	if (zz_steplist_delayed == NULL) {
		zz_commit_event = MAX_long;
		zz_commit_time  = TIME_inf;
	}

	queue_head (wsd, zz_steplist_delayed, ZZ_WSIT);

	// Update minimum intervals

	if (stepevent < zz_commit_event) zz_commit_event = stepevent;
	if (steptime < zz_commit_time) zz_commit_time = steptime;

	SETTIMER;
	return;
}

static  void    processEStepMode () {

/* ---------------------------- */
/* Exits globally the step mode */
/* ---------------------------- */

	ZZ_SIT     *ws, *wws;
	ZZ_WSIT    *wsd, *wwsd;

	CLEARTIMER;
	for (ws = zz_steplist; ws != NULL; ws = wws) {
		wws = ws->next;
		queue_out (ws);
		delete (ws);
	}

	for (wsd = zz_steplist_delayed; wsd != NULL; wsd = wwsd) {
		wwsd = wsd->next;
		queue_out (wsd);
		delete (wsd);
	}
	SETTIMER;
}

static  void    processEWindowPhrase () {

/* ----------------------------- */
/* Removes a window from display */
/* ----------------------------- */

	char    sname [256];
	int     i, mode, orgmode, stat;
	ZZ_WINDOW  *w;
	ZZ_SIT     *ws;
	ZZ_WSIT    *wsd;

	if (getOctet () != PH_TXT) {
		// Formal error
		while (getOctet () != PH_ESP);  // Skip the rest of the spurt
		sendError (DERR_ILS, "Standard name missing from EWN phrase");
		sendindicator = YES;
		return;
	}

	// Get the standard name of the window's object

	for (i = 0; i < 256; i++)
		if ((sname [i] = (char) (getOctet ())) == '\0') break;

	if (i >= 256) {
		// Standard name too long -- assume "illegal spurt" error
		while (getOctet () != PH_ESP);  // Skip the rest
		sendError (DERR_ILS,
		"Standard name in EWN phrase is longer than 256 characters");
		sendindicator = YES;
		return;
	}

	if (peekOctet () != PH_INM) {
		// Error -- mode indicator expected
		while (getOctet () != PH_ESP);  // Skip the rest
		sendError (DERR_ILS, "Mode indicator missing in a EWN phrase");
		sendindicator = YES;
		return;
	}

	if ((orgmode = (int) getInt ()) < 0) 
		// Illegal mode, make sure it is positive, though
		mode = MAX_int;
	else
		mode = orgmode;

	// Check if the window is station-relative

	if (peekOctet () == PH_INM) {
		// Station-relative window
		if ((stat = (int) getInt ()) < 0 || stat > NStations - 1) {
			// Illegal station -- skip to the end of the window
			// phrase and send the "illegal station" message
			sendOctet (PH_WCN);
			sendText (sname);
			sendInt (orgmode);
			sendInt (stat);
			sendError (DERR_ISN, "Illegal station number");
			sendOctet (PH_EPH);
			sendindicator = YES;
			return;
		}
	} else
		stat = NONE;    // Not station-relative

	// Find the window on the active list

	for (w = zz_windows; w != NULL; w = w->next) {
		if (w->mode != mode || w->stid != stat) continue;
		if (strcmp (w->obj->getSName (), sname) == 0) break;
	}

	if (w == NULL) {
		// Not found
		sendOctet (PH_WCN);
		sendText (sname);
		sendInt (orgmode);
		sendInt (stat);
		sendError (DERR_ONF, "Window not found");
		sendOctet (PH_EPH);
		sendindicator = YES;
		return;
	}

	zz_the_window = w;
	DisplayOpening = NO;
	DisplayClosing = YES;
	itemcount = 0;
	withinwindow = YES;
	if (w->frame != NULL) w->obj->displayOut (w->mode, w->stid);
	withinwindow = NO;

	CLEARTIMER;
	queue_out (w);          // Remove from the active list

	// Check if occurs on steplists

	for (ws = zz_steplist; ws != NULL; ws = ws->next)
		if (ws->win == w) {
			queue_out (ws);
			delete (ws);
			break;
		}

	for (wsd = zz_steplist_delayed; wsd != NULL; wsd = wsd->next)
		if (wsd->win == w) {
			queue_out (wsd);
			delete (wsd);
			break;
		}
	SETTIMER;
}

static  void    flushSupplist () {

/* ---------------------------------------- */
/* Sends the supplement lists to the server */
/* ---------------------------------------- */

	ZZ_SUPPLI  *sp, *spp;
	ZZ_SUPPLD  *sq, *sqq;
	int     cnt;

	if (zz_supplist == NULL && zz_suppldst == NULL) return;

	for (sp = zz_supplist, cnt = 0; sp != NULL; sp = sp->next) {
		if (cnt == 8) {
			SETTIMER; // Give them enough time
			cnt = 0;
		}
		// Addition
		sendOctet (PH_ADD);
		// The parents standard name
		sendText (sp->parent->getSName ());

		// The object's standard name
		sendText (stchp = sp->child->getSName ());

		// The object's base name
		if (strcmp (stchp, stchq = sp->child->getBName ()) == 0)
			sendText ("");
		else
			sendText (stchq);

		// The object's nickname
		if (sp->child->zz_nickname != NULL)
			sendText (sp->child->zz_nickname);
		else
			sendText ("");
	}

	for (sq = zz_suppldst, cnt = 0; sq != NULL; sq = sq->next) {
		if (cnt == 8) {
			SETTIMER; // Give them enough time
			cnt = 0;
		}
		// Deletion
		sendOctet (PH_REM);
		if (sq->identf == NONE)
			sendText (sq->tname);
		else
			sendText (form ("%s %1d", sq->tname, sq->identf));
		// Note: we assume that ports are never REMoved; their
		//       standard names are somewhat trickier
	}

	CLEARTIMER;
	// This cannot be interrupted
	for (sp = zz_supplist; sp != NULL; sp = spp) {
		spp = sp->next;
		delete (sp);
	}
	for (sq = zz_suppldst; sq != NULL; sq = sqq) {
		sqq = sq->next;
		delete (sq);
	}

	zz_suppcount = 0;
	zz_supplist = NULL;
	zz_suppldst = NULL;
	SETTIMER;
}
		
static  void    getResponse () {

/* ------------------------------------------------- */
/* Get the server's response and process the request */
/* ------------------------------------------------- */

	int     more           = YES,
		bt, oc;

	peekoctet = NONE;
	sendindicator = NO;

	while (more) {          // More stuff coming in

	 switch (oc = getOctet ()) {

	  int   d;

	  case  PH_ESP:         // The end of spurt

		more = NO;

	  case  PH_ATT:         // Attention -- ignore
	  case  PH_EPH:         // Spurious endphrase -- as above

		break;

	  case  PH_DIN:         // New display interval

		if ((d = peekOctet ()) == PH_INM) {

			if ((d = (int) getInt ()) < 0)
				sendError (DERR_NSI,
					"Negative display interval");
			else
				DisplayInterval = d;

		} else if (d == PH_EPH)
			DisplayInterval = DEFDINTVL;
		else {
			// Formal error in the spurt
			while (getOctet () != PH_ESP);
			sendError (DERR_ILS, "DIN phrase error");
		}

		// Echo the interval to the server

		sendOctet (PH_DIN);
		sendInt (DisplayInterval);
		sendindicator = YES;
		break;

	  case  PH_WND:         // A window request phrase

		zz_processWindowPhrase ();
		break;

	  case  PH_EWN:         // End-window phrase

		processEWindowPhrase ();
		break;

	  case  PH_ESM:         // Global exit from step mode

		processEStepMode ();
		break;

	  case  PH_END:         // Terminate display

		serviceEND ();
		return;

	  case  PH_EXI:         // Terminate simulation

		zz_max_Time = Time;
		break;

	  case  PH_RPL:         // A reply phrase

		{
			char    rbuf [256];
			int     i;

			if (peekOctet () != PH_TXT) {
				rplyptr = NULL;
				break;
			}
			getOctet ();
			for (i = 0; i < 255; i++) {
				if ((bt = getOctet ()) == '\0') break;
				rbuf [i] = bt;
			}
			rbuf [i] = '\0';
			rplyptr = form ("%s", rbuf);
			break;
		}

	  default:

		continue;       // Ignore it by now

	 }
	}

	if (sendindicator) {
		sendOctet (PH_ESP);
		SETTIMER;
		getResponse ();
	}

	// Note: nothing should follow this line, especially nothing including
	// send-receive requests, unless another timeout setjmp is issued.
}
		
void    refreshDisplay () {

/* ----------------- */
/* Process one cycle */
/* ----------------- */

	if (!DisplayActive) return;

	SYNCIO;         // Cancel asynchronous input from server

	if (setjmp (zz_tmoutenv)) {

		// Timeout exception
		CLEARTIMER;
		serviceEND ();
		return;
	}

	SETTIMER;       // To detect hangups

	if (initializing) {

		// The first time around -- reply with initial parameter setting

		initializing = NO;

		// BIG precision
		sendOctet (PH_BPR);
		sendInt (BIG_precision == TYPE_double ? 0 : BIG_precision * 31);

		// The number of stations
		sendOctet (PH_NST);
		sendInt (NStations);

		// Display interval
		sendOctet (PH_DIN);
		sendInt (DisplayInterval);

		// The initial ownership tree

		zz_sendTree (System);

		sendOctet (PH_ESP);

		getResponse ();    // Read and process the server's response

	}

	// Regular update

	flushSupplist ();
	withinwindow = YES;
	for (zz_the_window = zz_windows; zz_the_window != NULL;
		zz_the_window = zz_the_window->next) {

		// Set item number limit
		itemcount = zz_the_window->count;
		DisplayClosing = NO;
		DisplayOpening = (zz_the_window->flag & WF_OPENING);
		if (zz_the_window->flag & WF_STEPPED)
			zz_send_step_phrase = YES;
		// Start the phrase properly
		sendOctet (PH_WCN);
		sendText (zz_the_window->obj->getSName ());
		sendInt (zz_the_window->mode);
		sendInt (zz_the_window->stid);
		zz_dispfound = NO;
		zz_the_window->obj->displayOut (zz_the_window->mode,
			zz_the_window->stid);
		// Terminate the phrase
		sendOctet (PH_EPH);
		zz_the_window->flag = 0;        // Reset the status
		if (!zz_dispfound)
			sendError (DERR_NMO, "Nonexistent mode");
		sendOctet (PH_EPH);
	}
	withinwindow = NO;

	if (zz_send_step_phrase) {
		zz_send_step_phrase = NO;
		// Notify the server that some stepped windows have been
		// included
		sendOctet (PH_STP);
	}
	sendOctet (PH_ESP);

	getResponse ();         // Read and process the server's response

	CLEARTIMER;
	zz_events_to_display = DisplayInterval;
	ASYNCIO;                // Asynchronous input from the server
}

void    zz_DADD (ZZ_Object *pr, ZZ_Object *ob) {

/* ---------------------------------------------------------- */
/* Informs the server about a new object having been created. */
/* Note  that  once the simulation is started, only processes */
/* are allowed to create objects dynamically.                 */
/* ---------------------------------------------------------- */

	ZZ_SUPPLI  *sp;

	if (!DisplayActive) return;

	// Add to the supplement list

	if (zz_suppcount >= MAXSUPPL) {
		// Limit reached, flush the supplement list
		if (setjmp (zz_tmoutenv)) {
			// Timeout exception
			CLEARTIMER;
			serviceEND ();
			return;
		}

		SYNCIO;                 // Resume synchronous input from server
		SETTIMER;
		flushSupplist ();
		sendOctet (PH_ESP);     // Terminate properly
		getResponse ();
		CLEARTIMER;
		ASYNCIO;
	}

	sp = new ZZ_SUPPLI;

	sp->child = ob;
	sp->parent = pr;
	queue_head (sp, zz_supplist, ZZ_SUPPLI);
	zz_suppcount++;
}

void    zz_DREM (ZZ_Object *ob) {

/* ---------------------------------------------------------- */
/* Informs  the  server  about  the  removal of a displayable */
/* object                                                     */
/* ---------------------------------------------------------- */

	ZZ_SUPPLD  *sp;
	ZZ_SUPPLI  *si, *sj;
	ZZ_WINDOW  *w, *wq;
	ZZ_SIT     *ws, *wsq;
	ZZ_WSIT    *wws, *wwsq;

	if (!DisplayActive) return;

	// Check if the object occurs on the insert supplement list

	for (si = zz_supplist; si != NULL; si = sj) {
		sj = si->next;
		if (si->child == ob) {
			queue_out (si);
			zz_suppcount--;
			delete (si);
			return;
		}
	}

	// Add to the list of supplements

	if (zz_suppcount >= MAXSUPPL) {
		// Limit reached, flush the supplement list
		if (setjmp (zz_tmoutenv)) {
			// Timeout exception
			CLEARTIMER;
			serviceEND ();
			return;
		}

		SYNCIO;                 // Resume synchronous input from server
		SETTIMER;
		flushSupplist ();
		sendOctet (PH_ESP);     // Terminate properly
		getResponse ();
		CLEARTIMER;
		ASYNCIO;
	}

	sp = new ZZ_SUPPLD;

	sp->tname = ob->getTName ();
	sp->identf = ob->Id;
	queue_head (sp, zz_suppldst, ZZ_SUPPLD);
	zz_suppcount++;

	// Remove all windows related to this object

	for (ws = zz_steplist; ws != NULL; ws = wsq) {
		wsq = ws->next;
		if (ws->win->obj == ob) {
			queue_out (ws);
			delete (ws);
		}
	}

	for (wws = zz_steplist_delayed; wws != NULL; wws = wwsq) {
		wwsq = wws->next;
		if (wws->win->obj == ob) {
			queue_out (wws);
			delete (wws);
		}
	}

	for (w = zz_windows; w != NULL; w = wq) {
		wq = w->next;
		if (w->obj == ob) {
			queue_out (w);
			delete (w);
		}
	}

}

void    zz_dterminate () {

/* ------------------------------------ */
/* Informs the server about termination */
/* ------------------------------------ */

	if (setjmp (zz_tmoutenv)) {

		// Timeout exception
		CLEARTIMER;
		serviceEND ();
		return;
	}
	SYNCIO;
	SETTIMER;
	sendOctet (PH_EXI);
	refreshDisplay ();
}

static  int     stimeout;

char    *displayNote (char *t) {

/* ---------------------------------------------- */
/* Send a note to the server and await a response */
/* ---------------------------------------------- */

	if (!DisplayActive) return (NULL);

	if (setjmp (zz_tmoutenv)) {

		// Timeout exception
		CLEARTIMER;
		serviceEND ();
		return (NULL);
	}

	SYNCIO;
	SETTIMER;
	flushSupplist ();               // As usual

	stimeout = zz_ctimeout;         // Save default timeout
	zz_ctimeout = UTIMEOUT;         // Assume long timeout  

	SETTIMER;                       // Reset the timer
	sendOctet (PH_LCK);
	sendText (t);
	sendOctet (PH_ESP);
	rplyptr = NULL;
	getResponse ();
	CLEARTIMER;
	ASYNCIO;
	zz_ctimeout = stimeout;         // Resume previous timeout
	return (rplyptr);
}

void    display (long p) {

/* -------------------------------------------------- */
/* User-visible function to display an integer number */
/* -------------------------------------------------- */

	if (!withinwindow)
		excptn ("display (int): called from outside of expose");
	if (itemcount-- <= 0) return;
	sendInt (p);
}

#if     BIG_precision != 1

void    display (BIG p) {

/* --------------------------------------------- */
/* User-visible function to display a BIG number */
/* --------------------------------------------- */

	if (!withinwindow)
		excptn ("display (BIG): called from outside of expose");
	if (itemcount-- <= 0) return;
	sendBIG (p);
}

#endif

void    display (double p) {

/* -------------------------------------------------------- */
/* User-visible function to display a floating point number */
/* -------------------------------------------------------- */

	if (!withinwindow)
		excptn ("display (float): called from outside of expose");
	if (itemcount-- <= 0) return;
	sendFloat (p);
}

void    display (char *p) {

/* --------------------------------------------------- */
/* User-visible function to display a character string */
/* --------------------------------------------------- */

	if (!withinwindow)
		excptn ("display (text): called from outside of expose");
	if (itemcount-- <= 0) return;
	sendText (p);
}

void    display (char c) {

/* ----------------------------- */
/* To display a single character */
/* ----------------------------- */

	char    tt [2];

	if (!withinwindow)
		excptn ("display (char): called from outside of expose");

	tt[0] = c; tt[1] = '\0';
	sendText (tt);
};

static  void    sendScale (double xo, double xs, double yo, double ys) {

/* ---------------------------------------------- */
/* Send to the server the region scaling sequence */
/* ---------------------------------------------- */

	sendOctet (PH_SCA);
	sendFloat (xo);
	sendFloat (xs);
	sendFloat (yo);
	sendFloat (ys);
}

static  void    sendSegment (long att, int n, double *x, double *y) {

/* --------------------------------- */
/* Send to the server a line segment */
/* --------------------------------- */

	sendOctet (PH_SLI);
	if (att == NONE) {
		sendOctet (0);  // Defaults
		sendOctet (1);
	} else {
		sendOctet ((int)((att >> 7) & 0177));
		sendOctet ((int)(att & 0177));
	}

	while (n-- > 0) {
		sendFloat (*x++);
		sendFloat (*y++);
	}
}

void    display (double xo, double xs, double yo, double ys, long att,
			int n, double *x, double *y) {

/* ----------------------------------------------- */
/* Display a region consisting of a single segment */
/* ----------------------------------------------- */

	if (!withinwindow)
		excptn ("display (REG): called from outside of expose");
	if (itemcount-- <= 0) return;

	sendOctet (PH_REG);
	sendScale (xo, xs, yo, ys);
	sendSegment (att, n, x, y);
	sendOctet (PH_EPH);
}

void    display (long att, int n, double *x, double *y) {

/* -------------------------------------------------------- */
/* Display a region consisting of a single unscaled segment */
/* -------------------------------------------------------- */

	if (!withinwindow)
		excptn ("display (REG): called from outside of expose");
	if (itemcount-- <= 0) return;

	sendOctet (PH_REG);
	sendSegment (att, n, x, y);
	sendOctet (PH_EPH);
}

void    display (int n, double *x, double *y) {

/* ---------------------------------------------------------- */
/* Display  a  region consisting of a single unscaled segment */
/* with default attributes                                    */
/* ---------------------------------------------------------- */

	if (!withinwindow)
		excptn ("display (REG): called from outside of expose");
	if (itemcount-- <= 0) return;

	sendOctet (PH_REG);
	sendSegment (NONE, n, x, y);
	sendOctet (PH_EPH);
}

void    startRegion (double xo, double xs, double yo, double ys) {

/* ----------------------------------- */
/* Start a scaled multi-segment region */
/* ----------------------------------- */

	if (!withinwindow)
		excptn ("startRegion (...): called from outside of expose");
	if (itemcount <= 0) return;

	sendOctet (PH_REG);
	sendScale (xo, xs, yo, ys);
}

void    startRegion () {

/* -------------------------------------- */
/* Start an unscaled multi-segment region */
/* -------------------------------------- */

	if (!withinwindow)
		excptn ("startRegion: called from outside of expose");
	if (itemcount <= 0) return;

	sendOctet (PH_REG);
}

void    endRegion () {

/* -------------------------- */
/* End a multi-segment region */
/* -------------------------- */

	if (!withinwindow)
		excptn ("endRegion: called from outside of expose");
	if (itemcount-- <= 0) return;

	sendOctet (PH_EPH);
}

void    displaySegment (long att, int n, double *x, double *y) {

/* ------------------------------------------- */
/* Display a segment of a multi-segment region */
/* ------------------------------------------- */

	if (!withinwindow)
		excptn ("displaySegment: called from outside of expose");
	if (itemcount <= 0) return;
	sendSegment (att, n, x, y);
}

void    displaySegment (int n, double *x, double *y) {

/* ---------------------------------------------------------- */
/* Display   a  multi-segment  region  segment  with  default */
/* atributes                                                  */
/* ---------------------------------------------------------- */

	if (!withinwindow)
		excptn ("displaySegment: called from outside of expose");
	if (itemcount <= 0) return;
	sendSegment (NONE, n, x, y);
}

void    startSegment (long att) {

/* ----------------------------------------- */
/* Start a segment of a multi-segment region */
/* ----------------------------------------- */

	if (!withinwindow)
		excptn ("displaySegment: called from outside of expose");
	if (itemcount <= 0) return;

	sendOctet (PH_SLI);
	if (att == NONE) {
		sendOctet (0);  // Defaults
		sendOctet (1);
	} else {
		sendOctet ((int)((att >> 7) & 0177));
		sendOctet ((int)(att & 0177));
	}
}

void    displayPoint (double x, double y) {

/* ---------------------------------------------------------- */
/* Display   a  multi-segment  region  segment  with  default */
/* atributes                                                  */
/* ---------------------------------------------------------- */

	if (!withinwindow)
		excptn ("displaySegment: called from outside of expose");
	if (itemcount <= 0) return;
	sendFloat (x);
	sendFloat (y);
}

void    endSegment () {

/* ------------------- */
/* Terminate a segment */
/* ------------------- */

	// Truly void
	if (!withinwindow)
		excptn ("endSegment: called from outside of expose");
}
